from .colors import *
from .SunSpatial import *

import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.colors as mcolors
import matplotlib.tri as tri
from matplotlib.ticker import (MultipleLocator, AutoMinorLocator)

mpl.rcParams['font.size'] = 7


def draw_box(f, x, y, size, ax=None):
    w = 1
    h = 2
    box_full = [[w, -h], [w, h], [-w, h], [-w, -h]]
    ax.scatter([x], [y], marker=box_full, s=size, c="#cccccc", alpha=1, edgecolors='none', zorder=5)
    t = h*(-1 + 2*f)
    box_part = [[w, -h], [w, t], [-w, t], [-w, -h]]
    ax.scatter([x], [y], marker=box_part, s=size, c="#3a3a3a", alpha=1, edgecolors='none', zorder=5)


class SunPlotter:
    sunSpatialPlot: SunSpatial
    sunSpatialCalc: SunSpatial
    tris: np.ndarray
    levels: list
    labelPositionsE: list
    labelPositionsH: list

    def __init__(self, sunSpatialPlot: SunSpatial, sunSpatialCalc: SunSpatial, levels=[0., 1., 0.1, 10], 
                 labelPositionsE=[], labelPositionsH=[]
                 ):
        self.sunSpatialPlot = sunSpatialPlot
        self.sunSpatialCalc = sunSpatialCalc
        triangulation = tri.Triangulation(
            [e.hourAngle() for e in self.sunSpatialPlot.nodesE],
            [e.declination() for e in self.sunSpatialPlot.nodesE]
        )
        self.tris = triangulation.triangles
        self.levels = levels # start, finish, step, substeps
        self.labelPositionsE = labelPositionsE
        self.labelPositionsH = labelPositionsH

    def showEquatorial(self, etaF: typing.Callable, **kwargs):
        etaMin, etaMax, etaStep, etaSubsteps = kwargs.get('levels', self.levels)
        colorMap = kwargs.get('cmap', colorMapL)
        labelsFormat = kwargs.get('labelsFormat', '%0.1f')
        labelPositions = kwargs.get('labelPositions', self.labelPositionsE)
        cbarTitle = kwargs.get('cbarTitle', r'$\eta$')
        showWeights = kwargs.get('showWeights', False)

        pointsPlot = np.array([
            (e.hourAngle(), e.declination()) for e in self.sunSpatialPlot.nodesE
        ])
        xm = pointsPlot[:, 0]/angleHour
        ym = pointsPlot[:, 1]/degree
        zm = np.array([etaF(v) for v in self.sunSpatialPlot.nodesV])

        points = np.array([
            (e.hourAngle()/angleHour, e.declination()/degree) for e in self.sunSpatialCalc.nodesE
        ])
        values = np.array(self.sunSpatialCalc.nodesW.copy())
        if showWeights and len(values) > 0:
            values = np.fabs(values)
            vMax = np.max(values)
            values /= vMax
        else:
            showWeights = False

        etaSubstep = etaStep/etaSubsteps
        iMin = math.floor(etaMin/etaSubstep)    
        iMax = math.ceil(etaMax/etaSubstep)  
        levelsAll = etaSubstep*np.arange(iMin, iMax + 1)
        iMin = math.ceil(etaMin/etaStep)    
        iMax = math.floor(etaMax/etaStep) 
        levelsMajor = etaStep*np.arange(iMin, iMax + 1)
        ticksMajor = np.arange(etaMin, etaMax + 1e-6, etaStep)

        # figure
        figure, axes = plt.subplots(constrained_layout=True)
        figure.set_size_inches(3.5, 1.5)
        figure.set_dpi(150)

        # parts
        CS = axes.tricontourf(xm, ym, zm, levels=levelsAll, cmap=colorMap, zorder=-2, linestyles='solid')
        CS2 = axes.tricontour(xm, ym, zm, levels=CS.levels[::1], colors='#00000020', linewidths=0.3, linestyles='solid')
        CS3 = axes.tricontour(xm, ym, zm, levels=levelsMajor, colors='#00000050', linewidths=0.3,
                              linestyles='solid')

        axes.clabel(CS3, fmt=labelsFormat, fontsize=7, colors=colorAxes, manual=labelPositions)

        if showWeights:
            for q, w in zip(points, values):
                draw_box(w, q[0], q[1], 2*32, ax=axes)
        else:
            axes.scatter(points[:, 0], points[:, 1], s=[15], c="#000000", alpha=0.5, edgecolors='none', zorder=5)

        # frame
        axes.set_xlabel(r'Hour angle $\omega$, h')
        axes.set_xlim(-8, 8)
        axes.xaxis.set_major_locator(MultipleLocator(2))
        axes.xaxis.set_minor_locator(MultipleLocator(0.5))

        axes.set_ylabel(r'Declination $\delta$, deg')
        axes.set_ylim(-28, 28)
        if showWeights:
            axes.set_ylim(-30, 30)
        axes.yaxis.set_major_locator(MultipleLocator(10))
        axes.yaxis.set_minor_locator(MultipleLocator(2))

        axes.tick_params(which='both', direction='in', top=True, right=True, width=0.5, color=colorAxes)
        axes.tick_params(which='major', length=3)
        axes.tick_params(which='minor', length=1.5)

        for spine in axes.spines.values():
            spine.set_edgecolor(colorAxes)
            spine.set_linewidth(0.5)

        for v in [-6, 0, 6]:
            axes.axvline(x=v, c=colorGrid, ls='--', lw=0.5, zorder=-1)
        for v in [0]:
            axes.axhline(y=v, c=colorGrid, ls='--', lw=0.5, zorder=-1)

        # colorbar
        cbar = SunPlotter.addColorBar(figure, CS, 20, (), levelsMajor, cbarTitle)
        cbar.add_lines(CS3)

        return figure

    def showHorizontal(self, etaF: typing.Callable, **kwargs):
        etaMin, etaMax, etaStep, etaSubsteps = kwargs.get('levels', self.levels)
        colorMap = kwargs.get('cmap', colorMapL)
        labelsFormat = kwargs.get('labelsFormat', '%0.1f')
        labelPositions = kwargs.get('labelPositions', self.labelPositionsH)
        cbarTitle = kwargs.get('cbarTitle', r'$\eta$')
        showWeights = kwargs.get('showWeights', False)

        pointsPlot = np.array([
            (h.azimuth(), h.elevation()) for h in self.sunSpatialPlot.nodesH
        ])
        azs = pointsPlot[:, 0]/degree
        els = pointsPlot[:, 1]/degree
        zm = np.array([etaF(v) for v in self.sunSpatialPlot.nodesV])

        points = np.array([
            (h.azimuth()/degree, h.elevation()/degree) for h in self.sunSpatialCalc.nodesH
        ])
        values = np.array(self.sunSpatialCalc.nodesW.copy())
        if showWeights and len(values) > 0:
            values = np.fabs(values)
            vMax = np.max(values)
            values /= vMax
        else:
            showWeights = False

        etaSubstep = etaStep/etaSubsteps
        iMin = math.floor(etaMin/etaSubstep)    
        iMax = math.ceil(etaMax/etaSubstep)  
        levelsAll = etaSubstep*np.arange(iMin, iMax + 1)
        iMin = math.ceil(etaMin/etaStep)    
        iMax = math.floor(etaMax/etaStep) 
        levelsMajor = etaStep*np.arange(iMin, iMax + 1)
        ticksMajor = np.arange(etaMin, etaMax + 1e-6, etaStep)

        # figure
        figure, axes = plt.subplots(constrained_layout=True)

        figure.set_size_inches(3.5, 1.5)
        figure.set_dpi(150)

        # parts
        CS = axes.tricontourf(azs, els, self.tris, zm, levels=levelsAll, cmap=colorMap, zorder=-2, linestyles='solid')
        CS2 = axes.tricontour(azs, els, self.tris, zm, levels=CS.levels[::1], colors='#00000020', linewidths=0.3,
                              linestyles='solid')
        CS3 = axes.tricontour(azs, els, self.tris, zm, levels=levelsMajor, colors='#00000050', linewidths=0.3,
                              linestyles='solid')

        axes.clabel(CS3, fmt=labelsFormat, fontsize=7, colors=colorAxes, manual=labelPositions)

        if showWeights:
            for q, w in zip(points, values):
                draw_box(w, q[0], q[1], 2*32, ax=axes)
        else:
            axes.scatter(points[:, 0], points[:, 1], s=[15], c="#000000", alpha=0.5, edgecolors='none', zorder=5)

        # frame
        axes.set_xlabel(r'Azimuth $\gamma$, deg')
        axes.set_xlim(50, 360 - 50)
        axes.xaxis.set_major_locator(MultipleLocator(30))
        axes.xaxis.set_minor_locator(MultipleLocator(10))

        axes.set_ylabel(r'Elevation $\alpha$, deg')
        axes.set_ylim(-5, 95)
        axes.yaxis.set_major_locator(MultipleLocator(30))
        axes.yaxis.set_minor_locator(MultipleLocator(10))

        axes.tick_params(which='both', direction='in', top=True, right=True, width=0.5, color=colorAxes)
        axes.tick_params(which='major', length=3)
        axes.tick_params(which='minor', length=1.5)

        for spine in axes.spines.values():
            spine.set_edgecolor(colorAxes)
            spine.set_linewidth(0.5)

        colorD = '#B0B0B0FF'
        for v in np.arange(30, 360, 30):
            axes.axvline(x=v, c=colorD, ls='-', lw=0.25, zorder=-5)
        for v in np.arange(0, 90.1, 30):
            axes.axhline(y=v, c=colorD, ls='-', lw=0.25, zorder=-5)

        # colorbar
        cbar = SunPlotter.addColorBar(figure, CS, 20, (), levelsMajor, cbarTitle)
        cbar.add_lines(CS3)

        return figure

    @staticmethod
    def addColorBar(figure, CS, aspect, ticksLocs, ticksMajor, cbarTitle):
        cbar = figure.colorbar(CS, aspect=aspect)
        cbar.ax.set_xlabel(cbarTitle, labelpad=7)
        # cbar.ax.set_xlabel(r'${\rm W}/{\rm m}^2$', labelpad=7)
        # cbar.ax.set_ylabel('$\eta$', rotation=0)

        # cbar.ax.yaxis.set_major_locator(MultipleLocator(ticksLocs[0]))
        # cbar.ax.yaxis.set_minor_locator(MultipleLocator(ticksLocs[1]))
        cbar.set_ticks(ticksMajor)

        cbar.ax.tick_params(which='both', direction='in', top=True, right=True, width=0.5, color=colorAxes)
        cbar.ax.tick_params(which='major', length=3)
        cbar.ax.tick_params(which='minor', length=1.5)

        cbar.outline.set_linewidth(0.5)
        for spine in cbar.ax.spines.values():
            spine.set_edgecolor(colorAxes)
            spine.set_linewidth(0.5)

        return cbar



from .SunTemporal import *

def sampleInterval(a: float, b: float, rho: float, middle: bool = False) -> np.ndarray:
    """
    :param a: interval begin
    :param b: interval end
    :param rho: resolution
    :param middle: make point in the middle
    :return: points
    """
    if middle:
        n = 2*round((b - a)/(2*rho))
    else:
        n = round((b - a)/rho)
    return np.linspace(a, b, n + 1)


class SunSpatial:
    sc: SunCalculator
    nodesV: list  # vectors
    nodesH: list  # horizontal
    nodesE: list  # equatorial
    nodesW: list  # weights
    dims: list  # dimensions, used only for grid

    def __init__(self, sc: SunCalculator):
        self.sc = sc
        self.nodesW = []

    def sampleEquatorial(self, rho: float, grid: bool = False) -> None:
        """
        sample sun path in equatorial coordinates with a resolution
        :param rho: resolution
        :param grid: make stretched rectangular grid
        :return:
        """
        sc = self.sc

        ans = []
        deltas = sampleInterval(-sc.obliquity, sc.obliquity, rho)
        if grid:
            omegas = sampleInterval(-6.*angleHour, 6.*angleHour, rho, middle=True)
            self.dims = [len(omegas), len(deltas)]
            for delta in deltas:
                s = sc.findHourAngleMax(delta)/(6.*angleHour)
                for omega in omegas:
                    ans += [PointEquatorial(omega*s, delta)]
        else:
            self.dims = []
            for delta in deltas:
                omegaMax = sc.findHourAngleMax(delta)
                omegas = sampleInterval(-omegaMax, omegaMax, rho, middle=True)
                for omega in omegas:
                    ans += [PointEquatorial(omega, delta)]

        self.nodesE = ans
        self.nodesV = [sc.findVectorFromEquatorial(e) for e in self.nodesE]
        self.nodesH = [sc.findHorizontalFromVector(v) for v in self.nodesV]
        # clean -0
        for ae in self.nodesH:
            if abs(ae.elevation()) < 1e-16: ae.y = 0.

    def info(self):
        n = len(self.nodesE)
        print('points: {:n}'.format(n))

    def checkAccuracy(self, etaF: typing.Callable, etaF_ref: typing.Callable) -> None:
        """
        quick check of interpolation accuracy
        :param etaF: interpolated function
        :param etaF_ref: reference function
        :return:
        """
        pointsF = np.array([etaF(v) for v in self.nodesV])
        pointsF_ref = np.array([etaF_ref(v) for v in self.nodesV])
        delta_rms = np.sqrt(np.mean((pointsF - pointsF_ref) ** 2))
        delta_m = np.mean(pointsF - pointsF_ref)
        print('delta_rms = {:.3f}%'.format(delta_rms*100))
        print('delta_m = {:.3f}%'.format(delta_m*100))

    def integrate(self, values) -> float:
        """
        find annual integral
        :param values: values at sampling points or a function
        :return: integral
        """
        if isinstance(values, list):
            vs = values
        elif callable(values):
            vs = [values(v) for v in self.nodesV]
        temp = [v*o for v, o in zip(vs, self.nodesW)]
        return sum(temp)


import math
import mathutils as mu
import datetime
from datetime import timedelta

degree = math.pi/180
angleHour = 360*degree/24
arcMinute = degree/60
arcSecond = arcMinute/60


class Point:
    x: float
    y: float

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y


# class Vec2(mu.Vector):
#     @classmethod
#     def fromXY(cls, x: float, y: float):
#         return cls((x, y))
#
# Vec2.fromXY(1., 2.)


class PointHorizontal(Point):
    """
    horizontal coordinates
    """

    def __init__(self, azimuth: float = 0., elevation: float = 0.):
        Point.__init__(self, azimuth, elevation)

    def azimuth(self) -> float:
        return self.x

    def elevation(self) -> float:
        return self.y

    def __str__(self):
        return 'azimuth = {:.3f}d, elevation = {:.3f}d'.format(self.azimuth()/degree, self.elevation()/degree)


class PointEquatorial(Point):
    """
    equatorial coordinates
    """

    def __init__(self, hourAngle: float = 0., declination: float = 0.):
        Point.__init__(self, hourAngle, declination)

    def hourAngle(self) -> float:
        return self.x

    def declination(self) -> float:
        return self.y

    def __str__(self):
        return 'hour angle = {:.3f}d, declination = {:.3f}d'.format(self.hourAngle()/degree, self.declination()/degree)


class PointGeographic(Point):
    """
    geographic coordinates
    """

    def __init__(self, latitude: float = 0.*degree, longitude: float = 0.*degree):
        assert abs(latitude) <= 90.*degree
        assert abs(longitude) <= 180.*degree
        Point.__init__(self, longitude, latitude)

    def latitude(self) -> float:
        return self.y

    def longitude(self) -> float:
        return self.x

    def __str__(self):
        return 'latitude = {:.3f}d, longitude = {:.3f}d'.format(self.latitude()/degree, self.longitude()/degree)


class Location:
    """
    geographic location with name and time zone
    """
    name: str
    point: PointGeographic
    timeDelta: datetime.timedelta

    def __init__(self,
                 name: str = "",
                 point: PointGeographic = PointGeographic(),
                 timeDelta: timedelta = timedelta(hours=0)
                 ):
        self.name = name
        self.point = point
        self.timeDelta = timeDelta



import matplotlib.pyplot as plt
import matplotlib.cm as cm  # todo
import matplotlib.colors

colorMain = '#567CB5'
colorGrid = '#B0B0B0FF'
colorAxes = '#303030'

# https://matplotlib.org/stable/tutorials/colors/colormaps.html
# cMap = plt.cm.turbo #todo

# uniform rainbow from ParaView
colorList = [
    '#0561ff', '#056cf7', '#0577f0', '#0582e8', '#058bdf',
    '#0595d5', '#059ecb', '#05a6bf', '#05afb3', '#05b8a8',
    '#05c19a', '#05ca8c', '#05d47f', '#05dc6d', '#06e55c',
    '#04ed4b', '#46f327', '#7ef51c', '#a4f90c', '#c2fb09',
    '#e1fd06', '#ffff03', '#fff414', '#ffe826', '#ffdd37',
    '#ffd137', '#ffc437', '#ffb837', '#ffac37', '#ff9f37',
    '#ff9337', '#ff8537', '#ff7737', '#ff6837', '#fe5536',
    '#fc4230', '#fd2636', '#f21e40', '#e6144a', '#da0a54',
    '#cc0b5b', '#bd0c62', '#ae0d6a'
]


def makeColorMap(colors):
    ans = matplotlib.colors.LinearSegmentedColormap.from_list("", colors)
    ans.set_under(colors[0])
    ans.set_over(colors[-1])
    return ans


colorMapL = makeColorMap(colorList)
colorMapL_r = makeColorMap(colorList[::-1])  # reversed



from .SunCalculator import *

import suncalc

import csv
import urllib.request
from io import StringIO
import re


def irradianceIneichen(s: mu.Vector) -> float:
    """
    Ineichen model for clear sky irradiance (DNI)
    :param s:
    :return:
    """
    if s.z < 0.001:
        return 0.
    else:
        return 1618.*math.exp(-0.606/s.z ** 0.491)


class SunTemporal:
    sc: SunCalculator
    tStep: timedelta
    pointsV: list # sun positions
    pointsW: list # irradiance

    def __init__(self, sc: SunCalculator):
        self.sc = sc

    def sampleYear(self, year: int = 2021, tStep: timedelta = timedelta(minutes=60),
                   irradiance: typing.Callable = irradianceIneichen) -> None:
        """
        cache sun positions and irradiance with a fixed time step
        :param year: year
        :param tStep: time step
        :param irradiance: function for clear sky irradiance
        :return:
        """
        sc = self.sc
        self.tStep = tStep

        tz = datetime.timezone(sc.location.timeDelta)
        tA = datetime.datetime(year, 1, 1, 0, 0, 0, tzinfo=tz)
        tB = datetime.datetime(year + 1, 1, 1, 0, 0, 0, tzinfo=tz)

        self.pointsV = []
        latitude = sc.location.point.latitude()/degree
        longitude = sc.location.point.longitude()/degree
        t = tA
        while t < tB:
            sp = suncalc.get_position(t, longitude, latitude)
            h = PointHorizontal(sp['azimuth'] + math.pi, sp['altitude'])
            self.pointsV += [sc.findVectorFromHorizontal(h)]
            t += tStep

        self.pointsW = [irradiance(v) for v in self.pointsV]

    def readTMY(self, filename: str, tMid: float = -0.5) -> None:
        """
        read TMY weather file
        :param filename:
        :param tMid: add this offset (in units of time step) to get the center of interval relative to timeStamp
        :return:
        """
        sc = self.sc

        if filename.startswith("http"):
            response = urllib.request.urlopen(filename)
            file = StringIO(response.read().decode('utf-8'))
        else:
            file = open(filename, "r")

        reader = csv.reader(file, delimiter=',')

        header = next(reader)
        nLatitude = -1
        nLongitude = -1
        nTimeZone = -1
        for i, h in enumerate(header):
            if re.search('Latitude', h, re.IGNORECASE):
                nLatitude = i
            elif re.search('Longitude', h, re.IGNORECASE):
                nLongitude = i
            elif re.search('Time Zone', h, re.IGNORECASE):
                nTimeZone = i
        assert nLatitude >= 0
        assert nLongitude >= 0
        assert nTimeZone >= 0

        row = next(reader)
        latitude = float(row[nLatitude])
        longitude = float(row[nLongitude])
        timeZone = float(row[nTimeZone])

        sc.location.name = "TMY"
        sc.location.point = PointGeographic(latitude*degree, longitude*degree)
        sc.location.timeDelta = datetime.timedelta(hours=timeZone)

        header = next(reader)
        nYear = -1
        nMonth = -1
        nDay = -1
        nHour = -1
        nDNI = -1
        for i, h in enumerate(header):
            if re.search('Year', h, re.IGNORECASE):
                nYear = i
            elif re.search('Month', h, re.IGNORECASE):
                nMonth = i
            elif re.search('Day', h, re.IGNORECASE):
                nDay = i
            elif re.search('Hour', h, re.IGNORECASE):
                nHour = i
            elif re.search('DNI', h, re.IGNORECASE):
                nDNI = i
        assert nYear >= 0
        assert nMonth >= 0
        assert nDay >= 0
        assert nHour >= 0
        assert nDNI >= 0

        listT = []
        listW = []
        tz = datetime.timezone(sc.location.timeDelta)
        for row in reader:
            year = int(row[nYear])
            month = int(row[nMonth])
            day = int(row[nDay])
            hour = int(row[nHour])
            irradiance = float(row[nDNI])

            t = datetime.datetime(year, month, day, hour, 0, 0, tzinfo=tz)
            listT += [t]
            listW += [irradiance]

        self.tStep = listT[1] - listT[0]
        dt = tMid*self.tStep
        # print(dt)

        listV = []
        for t in listT:
            sp = suncalc.get_position(t + dt, longitude, latitude)
            h = PointHorizontal(sp['azimuth'] + math.pi, sp['altitude'])
            listV += [SunCalculator.findVectorFromHorizontal(h)]

        # a correct tMid should minimize wBelow
        wBelow = 0.
        wTotal = 0.
        for v, w in zip(listV, listW):
            wTotal += w
            if v.z < 0. and w > 0.:
                wBelow += w
        print('annual insolation (all elevations): {:.3f} kWh/m2'.format(wTotal/1000))
        print('annual insolation (negative elevations): {:.3f} kWh/m2'.format(wBelow/1000))

        self.pointsV = listV
        self.pointsW = listW

    def checkAccuracy(self, etaF: typing.Callable, etaRefF: typing.Callable) -> list:
        ansRMS = 0.
        ansM = 0.
        ansWM = 0.
        wTotal = 0.
        n = 0
        for v, w in zip(self.pointsV, self.pointsW):
            if v.z < 0.: continue
            diff = etaF(v) - etaRefF(v)
            ansRMS += diff ** 2
            ansM += diff
            ansWM += w*diff
            wTotal += w
            n += 1
        ansRMS = math.sqrt(ansRMS/n)
        ansM /= n
        ansWM /= wTotal
        print('delta_rms = {:.3f}%'.format(ansRMS*100))
        print('delta_m = {:.3f}%'.format(ansM*100))  # mean
        print('delta_wm = {:.3f}%'.format(ansWM*100))  # weighted mean
        return [ansRMS, ansM, ansWM]

    def integrate(self, etaF: typing.Callable) -> float:
        ans = 0.
        for v, w in zip(self.pointsV, self.pointsW):
            if v.z <= 0.: continue
            eta = etaF(v)
            ans += w*eta

        ans *= self.tStep.seconds/3600.
        return ans

    def average(self, etaF: typing.Callable) -> float:
        a = self.integrate(etaF)
        b = self.integrate(lambda x: 1.)
        return a/b


from .SunSpatial import *

from scipy import interpolate
from scipy.special import xlogy
from enum import Enum

import csv
import urllib.request
from io import StringIO


def readEtaRefCSV(filename: str) -> typing.Callable:
    """
    read simulation data from CSV
    :param filename:
    :return:
    """
    # read
    if filename.startswith("http"):
        response = urllib.request.urlopen(filename)
        file = StringIO(response.read().decode('utf-8'))
    else:
        file = open(filename, "r")

    reader = csv.reader(file, delimiter=',')  # quoting=csv.QUOTE_NONNUMERIC
    data = []
    for row in reader:
        data += [row]
    data = np.array(data)    

    # convert
    iAzimuth = np.where(data[0] == 'azimuth')[0][0]
    iElevation = np.where(data[0] == 'elevation')[0][0]
    iEfficiency = np.where(data[0] == 'efficiency')[0][0]
    assert iAzimuth >= 0
    assert iElevation >= 0
    assert iEfficiency >= 0
    
    ans = []
    for q in data[1:]:
        azimuth = float(q[iAzimuth])
        elevation = float(q[iElevation])
        eta = float(q[iEfficiency])
        ans.append([azimuth, elevation, eta])
    ans = np.array(ans)

    # interpolator
    gammas = np.unique(ans[:, 0])*degree
    alphas = np.unique(ans[:, 1])*degree
    etas = np.reshape(ans[:, 2], (len(gammas), len(alphas)))

    etaFuncTemp = interpolate.RectBivariateSpline(gammas, alphas, etas)

    def etaFunc(v: mu.Vector) -> float:
        h = SunCalculator.findHorizontalFromVector(v)
        return etaFuncTemp(h.azimuth(), h.elevation())[0, 0]

    return etaFunc

    
def readSolarPilot(filename: str) -> typing.Callable:
    """
    read simulation data from SolarPilot
    :param filename:
    :return:
    """
    # read
    if filename.startswith("http"):
        response = urllib.request.urlopen(filename)
        file = StringIO(response.read().decode('utf-8'))
    else:
        file = open(filename, "r")

    reader = csv.reader(file, delimiter=',')  # quoting=csv.QUOTE_NONNUMERIC
    data = []
    for row in reader:
        data += [row]
    data = np.array(data).T

    # convert
    iPowerIn = np.where(data[0] == 'Power incident on field')[0][0]
    iPowerOut = np.where(data[0] == 'Power absorbed by the receiver')[0][0]
    iAzimuth = np.where(data[0] == 'fluxsim: Solar azimuth angle (0=N)')[0][0]
    iElevation = np.where(data[0] == 'fluxsim: Solar elevation angle')[0][0]
    iAbsorption = np.where(data[0] == 'Absorption efficiency')[0][0]
    assert iPowerIn >= 0
    assert iPowerOut >= 0
    assert iAzimuth >= 0
    assert iElevation >= 0
    assert iAbsorption >= 0

    ans = []
    for q in data[2:-1]:
        azimuth = float(q[iAzimuth])
        elevation = float(q[iElevation])
        powerIn = float(q[iPowerIn])
        powerOut = float(q[iPowerOut])
        etaAbsorption = float(q[iAbsorption])/100
        eta = powerOut/etaAbsorption/powerIn
        ans.append([azimuth, elevation, eta])
    ans = np.array(ans)

    # interpolator
    gammas = np.unique(ans[:, 0])*degree
    alphas = np.unique(ans[:, 1])*degree
    etas = np.reshape(ans[:, 2], (len(gammas), len(alphas)))

    etaFuncTemp = interpolate.RectBivariateSpline(gammas, alphas, etas)

    def etaFunc(v: mu.Vector) -> float:
        h = SunCalculator.findHorizontalFromVector(v)
        return etaFuncTemp(h.azimuth(), h.elevation())[0, 0]

    return etaFunc


class SphericalRBF:
    nodes: list
    amplitudes: list
    sigma: float

    def kernel(self, ra: mu.Vector, rb: mu.Vector) -> float:
        arg = ra.dot(rb) - 1.
        arg /= self.sigma ** 2
        return math.exp(arg)

    def __init__(self, nodes: list, values: list, sigma: float = 20.*degree):
        self.sigma = sigma
        self.nodes = nodes
        matrix = np.array([[
            self.kernel(p, q)
            for q in self.nodes]
            for p in self.nodes])
        self.amplitudes = np.linalg.solve(matrix, values)
        # self.amplitudes = np.linalg.lstsq(matrix, values, rcond=None)[0]

    def __call__(self, v: mu.Vector) -> float:
        temp = [a*self.kernel(vn, v) for vn, a in zip(self.nodes, self.amplitudes)]
        return math.fsum(temp)


class InterpolationMethod:
    """
    method: str
        RBF-2D,
        RBF-3D
        triangular,
        B-spline

    kernel (for RBF-2D): str or callable
        'multiquadric': sqrt((r/self.epsilon)**2 + 1)
        'inverse': 1.0/sqrt((r/self.epsilon)**2 + 1)
        'gaussian': exp(-(r/self.epsilon)**2)
        'linear': r
        'thin_plate': r**2 * log(r)
        'cubic': r**3
        'quintic': r**5

    kernel (for RBF-3D): str or callable
        'polyharmonic':
        'gaussian'
        'gaussian-scipy'

    order:
        for triangular, polyharmonic and B-splines

    sigma: float
        shape parameter

    precondition:

    node: int
        for localized kernels
    """
    def __init__(self, method:str = 'RBF-3D', kernel='polyharmonic',
                 order=5, sigma=20*degree,
                 precondition=lambda s: 1., node=-1):
        self.method = method
        self.kernel = kernel
        self.order = order
        self.sigma = sigma
        self.precondition = precondition
        self.node = node

    def findWeights(self, sunTemporal: SunTemporal, sunSpatial: SunSpatial, etaRef: typing.Callable):
        overlaps = []
        n0 = self.node
        for n in range(len(sunSpatial.nodesV)):
            self.node = n
            etaL = MakeInterpolation(sunSpatial, etaRef, self)
            overlap = sunTemporal.integrate(etaL)
            overlaps.append(overlap)
        self.node = n0
        sunSpatial.nodesW = overlaps

    def checkDiff(self, sunTemporal: SunTemporal, sunSpatial: SunSpatial, etaRef: typing.Callable):
        integralRef = sunTemporal.integrate(etaRef)
        integralW = sunTemporal.integrate(lambda x: 1.)
        etaInterp = MakeInterpolation(sunSpatial, etaRef, self)
        integral = sunSpatial.integrate(etaInterp)
        return (integral - integralRef)/integralW


class InterpolationFormat(Enum):
    f_xy = 1
    f_xy_bound = 2
    f_xy_grid = 3
    f_xyz = 4
    f_v = 5


class MakeInterpolation:
    sc: SunCalculator
    method: InterpolationMethod
    interpolationFormat: InterpolationFormat
    etaInterpolated: typing.Callable

    def __init__(self, sunSpatial: SunSpatial, etaRef: typing.Callable,
                 method: InterpolationMethod = InterpolationMethod()):
        self.sc = sunSpatial.sc
        self.method = method
        im = self.method

        points = np.array([
            (e.hourAngle(), e.declination()) for e in sunSpatial.nodesE
        ])
        if im.node < 0:
            values = [etaRef(v)*im.precondition(v) for v in sunSpatial.nodesV]
        else:
            values = np.zeros(len(sunSpatial.nodesV))
            values[im.node] = im.precondition(sunSpatial.nodesV[im.node])

        if im.method == 'triangular' and im.order == 1:
            self.etaInterpolated = interpolate.LinearNDInterpolator(
                points, values, fill_value=0.
            )
            self.interpolationFormat = InterpolationFormat.f_xy_bound
        elif im.method == 'triangular' and im.order == 2:
            self.etaInterpolated = interpolate.CloughTocher2DInterpolator(
                points, values, fill_value=0.
            )
            self.interpolationFormat = InterpolationFormat.f_xy_bound
        elif im.method == 'RBF-2D':
            self.etaInterpolated = interpolate.Rbf(
                points[:, 0], points[:, 1], values, function=im.kernel
            )
            self.interpolationFormat = InterpolationFormat.f_xy
        elif im.method == 'RBF-3D':
            points = np.array([
                (v.x, v.y, v.z) for v in sunSpatial.nodesV
            ])
            if isinstance(im.kernel, str):
                if im.kernel == 'polyharmonic':
                    k = im.order
                    if k%2 == 0:
                        kernelt = lambda r: xlogy(r ** k, r)
                    else:
                        kernelt = lambda r: r ** k
                    self.etaInterpolated = interpolate.Rbf(
                        points[:, 0], points[:, 1], points[:, 2], values, function=kernelt
                    )
                    self.interpolationFormat = InterpolationFormat.f_xyz
                    return
                elif im.kernel == 'gaussian-scipy':
                    self.etaInterpolated = interpolate.Rbf(
                        points[:, 0], points[:, 1], points[:, 2], values,
                        norm=lambda ra, rb: np.dot(ra, rb) - 1.,
                        function=lambda s, r: np.exp(r/s.epsilon**2),
                        epsilon=im.sigma
                    )
                    self.interpolationFormat = InterpolationFormat.f_xyz
                    return
                elif im.kernel == 'gaussian':
                    self.etaInterpolated = SphericalRBF(sunSpatial.nodesV, values, im.sigma)
                    self.interpolationFormat = InterpolationFormat.f_v
                    return
            else:
                self.etaInterpolated = interpolate.Rbf(
                    points[:, 0], points[:, 1], points[:, 2], values, function=im.kernel
                )
                self.interpolationFormat = InterpolationFormat.f_xyz
                return
        elif im.method == 'B-spline':
            assert len(sunSpatial.dims) == 2
            omegas = np.linspace(-6*angleHour, 6*angleHour, sunSpatial.dims[0])
            deltas = np.linspace(-self.sc.obliquity, self.sc.obliquity, sunSpatial.dims[1])
            zs = np.reshape(np.array(values), (sunSpatial.dims[1], sunSpatial.dims[0])).T
            ks = im.order
            self.etaInterpolated = interpolate.RectBivariateSpline(omegas, deltas, zs, kx=ks[0], ky=ks[1])
            self.interpolationFormat = InterpolationFormat.f_xy_grid

    def __call__(self, v: mu.Vector) -> float:
        im = self.method
        pc = im.precondition(v)
        if self.interpolationFormat == InterpolationFormat.f_v:
            return self.etaInterpolated(v)/pc
        elif self.interpolationFormat == InterpolationFormat.f_xyz:
            return self.etaInterpolated(v.x, v.y, v.z)/pc
        elif self.interpolationFormat == InterpolationFormat.f_xy:
            e = self.sc.findEquatorialFromVector(v)
            return self.etaInterpolated(e.hourAngle(), e.declination())/pc
        elif self.interpolationFormat == InterpolationFormat.f_xy_bound:
            e = self.sc.findEquatorialFromVector(v)
            declination = np.clip(e.declination(), -self.sc.obliquity, self.sc.obliquity)
            return self.etaInterpolated(e.hourAngle(), declination)/pc
        elif self.interpolationFormat == InterpolationFormat.f_xy_grid:
            e = self.sc.findEquatorialFromVector(v)
            declination = np.clip(e.declination(), -self.sc.obliquity, self.sc.obliquity)
            s = 6*angleHour/self.sc.findHourAngleMax(declination)
            return self.etaInterpolated(s*e.hourAngle(), declination)[0][0]/pc

# Inverse Distance Weighted
# https://stackoverflow.com/questions/3104781/inverse-distance-weighted-idw-interpolation-with-python/3119544#3119544



from .coordinates import *

import numpy as np
import typing


class SunCalculator:
    obliquity: float
    location: Location

    def __init__(self, location: Location = Location()):
        self.obliquity = 23.4*degree
        self.location = location

    @staticmethod
    def findVectorFromHorizontal(h: PointHorizontal) -> mu.Vector:
        gamma = h.azimuth()
        cosGamma = math.cos(gamma)
        sinGamma = math.sin(gamma)
        alpha = h.elevation()
        cosAlpha = math.cos(alpha)
        sinAlpha = math.sin(alpha)
        return mu.Vector((
            sinGamma*cosAlpha,
            cosGamma*cosAlpha,
            sinAlpha
        ))

    @staticmethod
    def findHorizontalFromVector(v: mu.Vector) -> PointHorizontal:
        gamma = math.atan2(v.x, v.y)
        alpha = math.asin(v.z)
        if gamma < 0.:
            gamma += 2.*math.pi
        return PointHorizontal(gamma, alpha)

    def findVectorFromEquatorial(self, e: PointEquatorial) -> mu.Vector:
        phi = self.location.point.latitude()
        cosPhi = math.cos(phi)
        sinPhi = math.sin(phi)
        omega = e.hourAngle()
        cosOmega = math.cos(omega)
        sinOmega = math.sin(omega)
        delta = e.declination()
        cosDelta = math.cos(delta)
        sinDelta = math.sin(delta)
        return mu.Vector((
            -sinOmega*cosDelta,
            cosPhi*sinDelta - sinPhi*cosOmega*cosDelta,
            sinPhi*sinDelta + cosPhi*cosOmega*cosDelta
        ))

    def findEquatorialFromVector(self, v: mu.Vector) -> PointEquatorial:
        phi = self.location.point.latitude()
        cosPhi = math.cos(phi)
        sinPhi = math.sin(phi)
        omega = math.atan2(-v.x, -v.y*sinPhi + v.z*cosPhi)
        temp = np.clip(v.y*cosPhi + v.z*sinPhi, -1., 1.)
        delta = math.asin(temp)
        return PointEquatorial(omega, delta)

    def findEquatorialFromHorizontal(self, h: PointHorizontal) -> PointEquatorial:
        v = self.findVectorFromHorizontal(h)
        return self.findEquatorialFromVector(v)

    def findHorizontalFromEquatorial(self, e: PointEquatorial) -> PointHorizontal:
        v = self.findVectorFromEquatorial(e)
        return self.findHorizontalFromVector(v)

    def findHourAngleMax(self, declination: float) -> float:
        phi = self.location.point.latitude()
        temp = -math.tan(phi)*math.tan(declination)
        temp = np.clip(temp, -1., 1.)
        hourAngle = math.acos(temp)
        return hourAngle
